# ===== テキストファイル (UTF-8) から tankanji.dic を構築する =====
#
#  2024-03-06 版
#  http://hashimo.ddo.jp/
#

  use strict;
  use warnings;
  use utf8;

  my (%block, $adr, $blknam, $yo, $file);

  binmode STDOUT, ':raw:encoding(cp932)';
  binmode STDERR, ':raw:encoding(cp932)';
  binmode STDIN, ':raw:encoding(cp932)';

  # %block に読み込む

  foreach $file (@ARGV) {
    my ($buf);
    open FP, '<:utf8', $file;
    while ($buf = <FP>) {
      chomp $buf;
      if ($buf =~ /^(\S+)\t(\S+)$/) {
        my (@yomi, $y, $kanji, $l, $i);
        $y = $1;
        $kanji = $2;
        $l = -1;
        for ($i = 0; $i < length($y); $i ++) {
          $l ++;
          $yomi[$l] .= substr($y, $i, 1);
        }
        $y = '';
        $block{''}{'n'}{$yomi[0]} = 1;
        for ($i = 0; $i <= $l; $i ++) {
          if (&chkcode($yomi[$i]) == -1) {
            die '読みが半角または、基本多言語面を逸脱する文字を含んでいます';
          }
          if ($yomi[$i] =~ /\p{sc=Katakana}/) {  # ただの \p{Katakana} では
                                                 # Perl v5.26 以降で長音符にも
                                                 # マッチしてしまうので改訂
            die '読みがカタカナを含んでいます';
          }
          unless ($yomi[$i] =~ /\p{Hiragana}/) {
            if (index('！″＃＄％＆′（）＊＋，－．／０１２３４５６７８９' .
                '：；＜＝＞？＠ＡＢＣＤＥＦＧＨＩＪＫＬＭＮＯＰＱＲＳＴＵ' .
                'ＶＷＸＹＺ［￥］＾＿｀ａｂｃｄｅｆｇｈｉｊｋｌｍｎｏｐｑ' .
                'ｒｓｔｕｖｗｘｙｚ｛｜｝～。「」、・ー', $yomi[$i]) < 0) {
              die '読みに使用不能な文字が含まれています';
            }
          }
          $y .= $yomi[$i];
          if ($i < $l) {
            $block{$y}{'n'}{$yomi[$i + 1]} = 1;
          }
          else {
            my ($u);
            $u = &chkcode($kanji);
            if ($u == -1) {
              die "変換後の文字が一文字でないか、半角文字であるか、\n" .
                "基本多言語面を逸脱する文字です";
            }
            if (defined $block{$y}{'k'}{$kanji}) {
              if ($block{$y}{'k'}{$kanji} == 1) {
                die '二重登録です';
              }
            }
            $block{$y}{'k'}{$kanji} = 1;
          }
        }
      }
      else {
        die '行の構成が不正です';
      }
    }
    close FP;
  }

  # $block{$blknam}{'a'} を設定

  $adr = 0;
  foreach $blknam (sort keys %block) {
    my ($sz, $nk, $ny);
    $nk = keys %{$block{$blknam}{'k'}};
    if ($blknam ne '') {
      die '内部エラーです' if ($adr == 0);
      $sz = 0;
    }
    else {
      die '内部エラーです' if ($adr != 0);
      die '内部エラーです' if ($nk != 0);
      $sz = 14;
    }
    $block{$blknam}{'a'} = $adr;
    $ny = keys %{$block{$blknam}{'n'}};
    $sz += $nk * 2 + 2 + $ny * 5 + 2;
    $adr += $sz;
  }

  # 戻るポインタが無いかチェック

  foreach $blknam (keys %block) {
    my ($y);
    foreach $y (sort keys %{$block{$blknam}{'n'}}) {
      if ($block{$blknam}{'a'} >= $block{$blknam . $y}{'a'}) {
        die '内部エラーです';
      }
    }
  }

  # tankanji.dic を作成

  open FP, '> tankanji.dic';
  binmode FP;
  print FP 'TANKANJI20070404';
  foreach $yo (sort keys %{$block{''}{'n'}}) {
    print FP &mki(&chkcode($yo));
    print FP &mk3($block{$yo}{'a'});
  }
  print FP "\0\0";
  foreach $blknam (sort keys %block) {
    my ($kanji);
    next if ($blknam eq '');
    foreach $kanji (sort keys %{$block{$blknam}{'k'}}) {
      print FP &mki(&chkcode($kanji));
    }
    print FP "\0\0";
    foreach $yo (sort keys %{$block{$blknam}{'n'}}) {
      print FP &mki(&chkcode($yo));
      print FP &mk3($block{$blknam . $yo}{'a'});
    }
    print FP "\0\0";
  }
  close FP;

sub mki {
  # 数値を2バイト文字列(BigEndian)にする
  my ($dat) = @_;
  $dat = 0 if ($dat eq '');
  $dat &= 0xffff;
  return sprintf("%c%c", int($dat / 256), $dat % 256);
}

sub mk3 {
  # 数値を3バイト文字列(BigEndian)にする
  my ($dat) = @_;
  my ($i, $r);
  $dat = 0 if ($dat eq '');
  $dat &= 0xffffff;
  $r = '';
  for ($i = 256 * 256; $i >= 1; $i /= 256) {
    my ($wk);
    $wk = int($dat / $i);
    $r .= sprintf("%c", $wk);
    $dat = $dat % $i;
  }
  return $r;
}

sub chkcode {
  # 引数の文字に対するUnicode(数値)を返す
  # 2文字以上あったり0文字であったら -1 を返す
  # BMP外だったら -1 を返す
  # 半角だったら -1 を返す

  my ($str) = @_;
  my ($u);

  return -1 if ($str eq '');
  return -1 if (length($str) > 1);
  $u = ord($str);
  return -1 if ($u < 0x80);
  return -1 if ($u > 0xffff);
  return -1 if (0xff60 <= $u and $u < 0xffa0); # これも半角
  return $u;
}

__END__
